﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace WineFermentationAITool.LibFiles
{
    public class Chromosome
    {
        public double[,] W12,
                        W23;
        public double[] HBias,
                        OBias;
        public double Fitness; 
     
        public Chromosome() { }
        public Chromosome(Chromosome c)
        {
            W12 = new double[c.W12.GetLength(0), c.W12.GetLength(1)];
            W23 = new double[c.W23.GetLength(0), c.W23.GetLength(1)];
            HBias = new double[c.HBias.Length];
            OBias = new double[c.OBias.Length];
            Array.Copy(c.W12, this.W12, c.W12.Length);
            Array.Copy(c.W23, this.W23, c.W23.Length);
            Array.Copy(c.HBias, this.HBias, c.HBias.Length);
            Array.Copy(c.OBias, this.OBias, c.OBias.Length);
            Fitness = 0;
        }

        public Chromosome(int N1, int N2, int N3)
        {
            W12 = new double[N2, N1];
            W23 = new double[N3, N2];
            HBias = new double[N2];
            OBias = new double[N3];
            Fitness = 0;
            double mean = 0;
            double stdDist = Math.Sqrt(2.0 / (N1 + 1));

            for (int h = 0; h < N2; h++)
            {
                for (int i = 0; i < N1; i++)
                {
                    W12[h, i] = MathUtils.GetRandGaussianVal(mean, stdDist);
                }
                HBias[h] = 0;
            }

            stdDist = Math.Sqrt(2.0 / (N2 + 1));
            for (int o = 0; o < N3; o++)
            {
                for (int h = 0; h < N2; h++)
                {
                    W23[o, h] = MathUtils.GetRandGaussianVal(mean, stdDist);
                }
                OBias[o] = 0;
            }
        }
    
    }
}
