﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.IO;
using Newtonsoft.Json;

namespace WineFermentationAITool.LibFiles
{
    public class GA
    {
        List<Chromosome> Population;
        List<Chromosome> OrderedPopulation;
        NeuralNetwork GANeuralNetwork;
        Chromosome Parent1,
                   Parent2;
        Chromosome Offspring1,
                   Offspring2;
        public int N2;
        int PopulationSize,
            N1,
            N3,
            MaxGenerations,
            MaxEpochsGA;
        double Pmutation0,
               Pcrossover0,
               Beta;
        bool IsElitist;
        List<Chromosome> IndividualsFitness;

        public bool IsType1;
        public bool IsSigmoid;
        public double LearningRate;

        public GA(int populationSize, int maxEpochsGA, int maxGenerations, double pCrossover, double pMutation, double beta, bool isElitist, int N1, int N2, int N3, double learningRate, bool isSigmoidFunction, bool isType1, bool[] isTrainingSet, List<double[,]> dataSets, string weightsAction, string wFileName)
        {
            this.PopulationSize = populationSize;
            this.N1 = N1;
            this.N2 = N2;
            this.N3 = N3;
            this.MaxGenerations = maxGenerations;
            this.MaxEpochsGA = maxEpochsGA;
            this.Pmutation0 = pMutation;
            this.Pcrossover0 = pCrossover;
            this.Beta = beta;
            this.IsElitist = isElitist;
            this.IsType1 = isType1;
            this.IsSigmoid = isSigmoidFunction;
            IndividualsFitness = new List<Chromosome>();

            InitializePopulation(weightsAction, wFileName);
            //create a network to be used as fitness function
            this.GANeuralNetwork = new NeuralNetwork(IsType1, this.N2 - this.N1, learningRate * 10, IsSigmoid, "xavier", MaxEpochsGA, isTrainingSet, dataSets);
        }

        void InitializePopulation(string weightsAction, string wFileName)
        {
            Population = new List<Chromosome>();

            if (weightsAction == Constants.LOAD)
            {
                using (StreamReader sr = new StreamReader(wFileName))
                {
                    IsType1 = Convert.ToBoolean(sr.ReadLine().Split(':')[1]);
                    N2 = Convert.ToInt32(sr.ReadLine().Split(':')[1]);
                    IsSigmoid = Convert.ToBoolean(sr.ReadLine().Split(':')[1]);
                    PopulationSize = Convert.ToInt32(sr.ReadLine().Split(':')[1]);

                    for (int i = 0; i < PopulationSize; i++)
                    {
                        var chromosomeString = sr.ReadLine();
                        var chromosome = chromosomeString == null ? new() : JsonConvert.DeserializeObject<Chromosome>(chromosomeString);
                        Population.Add(chromosome);
                    }
                }
            }
            else
            {
                for (int i = 0; i < PopulationSize; i++)
                {
                    var chromosome = new Chromosome(N1, N2, N3);
                    Population.Add(chromosome);
                }
            }

            if (weightsAction == Constants.SAVE)
            {
                using (StreamWriter sw = new StreamWriter(wFileName))
                {
                    sw.WriteLine("IsType1:" + Convert.ToString(IsType1));
                    sw.WriteLine("N2:" + Convert.ToString(N2));
                    sw.WriteLine("IsSigmoid:" + Convert.ToString(IsSigmoid));
                    sw.WriteLine("PopulationSize:" + Convert.ToString(PopulationSize));
                    foreach (var chromosome in Population)
                    {
                        var chromosomeString = JsonConvert.SerializeObject(chromosome);
                        sw.WriteLine(chromosomeString);
                    }
                }
            }
        }
        
        void SetWeights(Chromosome Individual)
        {
            double[] W;

            for (int h = 0; h < N2; h++)
            {
                W = Individual.W12.Cast<double>().ToArray();
                Array.Copy(W, h * N1, GANeuralNetwork.HidLayer.Neurons[h].W, 0, N1);
                GANeuralNetwork.HidLayer.Neurons[h].Bias = Individual.HBias[h];
            }
            
            for (int o = 0; o < N3; o++)
            {
                W = Individual.W23.Cast<double>().ToArray();
                Array.Copy(W, o * N2, GANeuralNetwork.OutLayer.Neurons[o].W, 0, N2);
                GANeuralNetwork.OutLayer.Neurons[o].Bias = Individual.HBias[o];
            }
        }

        public void getIndividualsFitness(Chromosome Individual)
        {
            SetWeights(Individual);
            int CurrentIteration = 0;
            while (CurrentIteration++ < MaxEpochsGA)
            {
                Individual.Fitness = GANeuralNetwork.Backpropagation();
            }
            IndividualsFitness.Add(Individual);
        }
        void EvaluateIndividuals()
        {
            var threads_counter = 0;
            List<Task> tasks = new ();

            //check fitness for each individual
            foreach (var Individual in Population)
            {
                /* run a  maximum of  8 threads */
                if (threads_counter == Constants.MAX_NUM_OF_THREADS)
                { 
                    /* Wait for all the 8 tasks to finish and clear them from the list */
                    Task.WaitAll(tasks.ToArray());
                    tasks.Clear();
                    threads_counter = 0;
                }

                tasks.Add(Task.Factory.StartNew(() => getIndividualsFitness(Individual)));
            }

            Task.WaitAll(tasks.ToArray());
            tasks.Clear();
            OrderedPopulation = new List<Chromosome>();
            //sort individuals by fitness
            OrderedPopulation = IndividualsFitness.OrderBy(o=>o.Fitness).ToList();
        }
        void SelectParents()
        {
            if (IsElitist)
            {
                Parent1 = new Chromosome(OrderedPopulation.First());
                Parent2 = new Chromosome(OrderedPopulation.ElementAt(1));
            }
            else
            {
                List<Chromosome> RandIndividuals = Population.OrderBy(individual => MathUtils.Rand.Next()).Take(5).ToList();
                RandIndividuals = RandIndividuals.OrderBy(o => o.Fitness).ToList();
                Parent1 = new Chromosome(RandIndividuals.First());
                Parent2 = new Chromosome(RandIndividuals.ElementAt(1));
            }
        }
        void Crossover()
        {
            bool isExchange;
            double probability;
            //use uniform crossover

            for (int h = 0; h < N2; h++)
            {
                for (int i = 0; i < N1; i++)
                {
                    probability = MathUtils.GetRandProbability();
                    if (probability < Pcrossover0)
                    {
                        isExchange = Convert.ToBoolean(MathUtils.Rand.Next(2));
                        if (isExchange)
                        {
                        MathUtils.Swap(ref Parent1.W12[h, i], ref Parent2.W12[h, i]);
                        }
                    }
                }
                probability = MathUtils.GetRandProbability();
                if (probability < Pcrossover0)
                {
                    isExchange = Convert.ToBoolean(MathUtils.Rand.Next(2));
                    if (isExchange)
                    {
                        MathUtils.Swap(ref Parent1.HBias[h], ref Parent2.HBias[h]);
                    }
                }
            }

            for (int o = 0; o < N3; o++)
            {
                for (int h = 0; h < N2; h++)
                {
                    probability = MathUtils.GetRandProbability();
                    if (probability < Pcrossover0)
                    {
                        isExchange = Convert.ToBoolean(MathUtils.Rand.Next(2));
                        if (isExchange)
                        {
                        MathUtils.Swap(ref Parent1.W23[o, h], ref Parent2.W23[o, h]);
                        }
                    }
                }
                probability = MathUtils.GetRandProbability();
                if (probability < Pcrossover0)
                {
                    isExchange = Convert.ToBoolean(MathUtils.Rand.Next(2));
                    if (isExchange)
                    {
                        MathUtils.Swap(ref Parent1.OBias[o], ref Parent2.OBias[o]);
                    }
                }
            }
            Offspring1 = new Chromosome(Parent1);
            Offspring2 = new Chromosome(Parent2);
        }
        void MutateGene(ref double gene, double pMutation, int prevLayer)
        {
            double probability;

            probability = MathUtils.GetRandProbability();
            if (probability < pMutation)
            {
                if (prevLayer == 0)
                {
                    gene = 0;
                }
                else
                {
                    gene = MathUtils.GetRandGaussianVal(0, Math.Sqrt(2.0 / (prevLayer + 1)));
                }
            }
        }
        void Mutation(int generation)
        {
            double pMutation = Pmutation0 * Math.Pow(Math.E, (1 - generation) * Beta);

            for (int h = 0; h < N2; h++)
            {
                for (int i = 0; i < N1; i++)
                {
                    MutateGene(ref Offspring1.W12[h, i], pMutation, N1);
                    MutateGene(ref Offspring2.W12[h, i], pMutation, N1);
                }
                MutateGene(ref Offspring1.HBias[h], pMutation, 0);
                MutateGene(ref Offspring2.HBias[h], pMutation, 0);
            }

            for (int o = 0; o < N3; o++)
            {
                for (int h = 0; h < N2; h++)
                {
                    MutateGene(ref Offspring1.W23[o, h], pMutation, N2);
                    MutateGene(ref Offspring2.W23[o, h], pMutation, N2);
                }
                MutateGene(ref Offspring1.OBias[o], pMutation, 0);
                MutateGene(ref Offspring2.OBias[o], pMutation, 0);
            }
        }
        void EvaluateOffsprings()
        {
            SetWeights(Offspring1);
            SetWeights(Offspring2);
            Offspring1.Fitness = double.MaxValue;
            Offspring2.Fitness = double.MaxValue;

            int CurrentIteration = 0;
            while (CurrentIteration++ < MaxEpochsGA)
            {
                Offspring1.Fitness = GANeuralNetwork.Backpropagation();
                Offspring2.Fitness = GANeuralNetwork.Backpropagation();
            }

            foreach (var Individual in OrderedPopulation)
            {
                if (Offspring1.Fitness < Individual.Fitness)
                {
                    OrderedPopulation.Remove(OrderedPopulation.Last());
                    OrderedPopulation.Add(Offspring1);
                    break;
                }
            }
            OrderedPopulation = OrderedPopulation.OrderBy(o => o.Fitness).ToList();

            foreach (var Individual in OrderedPopulation)
            {
                if (Offspring2.Fitness < Individual.Fitness)
                {
                    OrderedPopulation.Remove(OrderedPopulation.Last());
                    OrderedPopulation.Add(Offspring2);
                    break;
                }
            }
            OrderedPopulation = OrderedPopulation.OrderBy(o => o.Fitness).ToList();
        }
        public Chromosome GAOptimization()
        {
            int currGeneration = 0;
            EvaluateIndividuals();
            while (currGeneration < MaxGenerations)
            {
                SelectParents();
                Crossover();
                Mutation(currGeneration);
                EvaluateOffsprings();
                currGeneration++;
            }
            //take best individual
            return OrderedPopulation.First();
        }
    }
}
