﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.InteropServices;
using OfficeOpenXml;
using System.IO;
using System.Security.Cryptography;

namespace WineFermentationAITool.LibFiles
{

    public class NeuralNetwork
    {
        int N1,
            N2,
            N3;
        double[] Expected;
        double[,] TrainingSet,
                  TestingSet;
        public Layer InLayer,
                     HidLayer,
                     OutLayer;
        double LearningRate,
               TrainingError;
        int MaxEpochs;
        bool IsType1;
        bool[] IsTrainingSet;
        string InitializationType;

        List<double[,]> DataSets;
        GA MyGA;
        PSO MyPSO;
        bool IsSigmoidFunction;

        public NeuralNetwork(bool isType1, int N2, bool isSigmoidFunction, double[,] W12, double[,] W23, double[] hBias, double[] oBias)
        {
            this.IsType1 = isType1;
            if (IsType1)
            {
                this.N1 = Constants.cInLayer;
            }
            else
            {
                this.N1 = Constants.cInLayer + 2;
            }
            this.N2 = N2;
            this.N3 = Constants.cOutLayer;
            this.IsSigmoidFunction = isSigmoidFunction;
            this.InLayer = new Layer(N1, 1, true);
            this.HidLayer = new Layer(N2, N1, W12, hBias);
            this.OutLayer = new Layer(N3, N2, W23, oBias);
        }
        public NeuralNetwork(bool isType1, int n, double learningRate, bool isSigmoidFunction, string initializationType, int maxEpochs = 0, bool[] isTrainingSet = null, List<double[,]> dataSets = null)
        {
            this.IsType1 = isType1;
            if (IsType1)
            {
                this.N1 = Constants.cInLayer;
            }
            else
            {
                this.N1 = Constants.cInLayer + 2;
            }
            this.N2 = N1 + n;
            this.N3 = Constants.cOutLayer;
            this.Expected = new double[N3];
            this.LearningRate = learningRate;
            this.MaxEpochs = maxEpochs;
            this.IsSigmoidFunction = isSigmoidFunction;
            this.InitializationType = initializationType;

            this.InLayer = new Layer(N1, 1, true);
            if (initializationType == Constants.XAVIER)
            {
                this.HidLayer = new Layer(N2, N1);
                this.OutLayer = new Layer(N3, N2);
            }
            DataSets = new List<double[,]>();
            this.DataSets = dataSets;
            this.IsTrainingSet = isTrainingSet;
            SetTrainingAndTestingSet();
        }

        public void SetGAParams(int populationSize, int maxEpochsGA, int maxGenerations, double pCrossover, double pMutation, double Beta, bool isElitist, bool[] isTrainingSet, List<double[,]> dataSets, string weightsAction, string wFileName)
        {
            MyGA = new GA(populationSize,maxEpochsGA, maxGenerations, pCrossover, pMutation, Beta, isElitist, N1, N2, N3, LearningRate, IsSigmoidFunction, IsType1, isTrainingSet, dataSets, weightsAction, wFileName);
            Chromosome BestIndividual = MyGA.GAOptimization();
            if (weightsAction == Constants.LOAD)
            {
                IsType1 = MyGA.IsType1;
                N2 = MyGA.N2;
                IsSigmoidFunction = MyGA.IsSigmoid;
            }
            this.HidLayer = new Layer(N2, N1, BestIndividual.W12, BestIndividual.HBias);
            this.OutLayer = new Layer(N3, N2, BestIndividual.W23, BestIndividual.OBias);
        }

        public void SetPSOParams(int numOfParticles, double c1, double c2, double w, int numOfIterations, int maxEpochsPSO,  bool[] isTrainingSet, List<double[,]> dataSets, string weightsAction, string wFileName)
        {
            MyPSO = new PSO(numOfParticles, c1, c2, w, numOfIterations, maxEpochsPSO, N1, N2, N3, LearningRate, IsSigmoidFunction, IsType1, isTrainingSet, dataSets, weightsAction, wFileName);
            Chromosome BestIndividual = MyPSO.PSOOptimization();
            if (weightsAction == Constants.LOAD)
            {
                IsType1 = MyPSO.IsType1;
                N2 = MyPSO.N2;
                IsSigmoidFunction = MyPSO.IsSigmoid;
            }
            this.HidLayer = new Layer(N2, N1, BestIndividual.W12, BestIndividual.HBias);
            this.OutLayer = new Layer(N3, N2, BestIndividual.W23, BestIndividual.OBias);
        }

        public Chromosome GetNeuralNetworkConfiguration()
        {
            double[] W;
            Chromosome config = new();
            config.W12 = new double[N2, N1];
            config.W23 = new double[N3, N2];
            config.HBias = new double[N2];
            config.OBias = new double[N3];

            for (int h = 0; h < N2; h++)
            {
                for (int i = 0; i < N1; i++)
                {
                    config.W12[h, i] = HidLayer.Neurons[h].W[i];
                }
                config.HBias[h] = HidLayer.Neurons[h].Bias;
            }

            for (int o = 0; o < N3; o++)
            {
                for (int h = 0; h < N2; h++)
                {
                    config.W23[o, h] = OutLayer.Neurons[o].W[h];
                }
                config.OBias[o] = OutLayer.Neurons[o].Bias;
            }

            return config;
        }
        public double[,] GetTestingSet()
        {
            return TestingSet;
        }

        private void SetTrainingAndTestingSet()
        {
            int numOfSets = DataSets.Count();
            int currSet = 0;
            int prevTrainingSetLength = 0,
                prevTestingSetLength = 0;
            int testingSetNumRows = 0,
                trainingSetNumRows = 0,
                numCols = 0;

            foreach (var isTrainingSet in IsTrainingSet)
            {
                if (isTrainingSet)
                {
                    trainingSetNumRows += (DataSets.ElementAt(currSet).GetLength(0));
                }
                else
                {
                    testingSetNumRows += (DataSets.ElementAt(currSet).GetLength(0));
                }
                if (currSet == 0)
                {
                    numCols = DataSets.ElementAt(currSet).GetLength(1);
                }
                currSet++;
            }
            TrainingSet = new double[trainingSetNumRows, numCols];
            TestingSet = new double[testingSetNumRows, numCols];
            currSet = 0;
            foreach (var isTrainingSet in IsTrainingSet)
            {
                if (isTrainingSet)
                {
                    Array.Copy(DataSets.ElementAt(currSet), 0, TrainingSet, prevTrainingSetLength, DataSets.ElementAt(currSet).Length);
                    prevTrainingSetLength += DataSets.ElementAt(currSet).Length;
                }
                else
                {
                    Array.Copy(DataSets.ElementAt(currSet), 0, TestingSet, prevTestingSetLength, DataSets.ElementAt(currSet).Length);
                    prevTestingSetLength += DataSets.ElementAt(currSet).Length;
                }
                currSet++;
            }            
        }
        private void SetData(int index, bool isTrainingSet)
        {
            if (isTrainingSet)
            {
                for (int i = 0; i < N1; i++)
                {
                    InLayer.Neurons[i].Out = TrainingSet[index, i];
                }

                for (int o = 0; o < N3; o++)
                {
                    Expected[o] = TrainingSet[index, N1 + o];
                }
            }
            else
            {
                for (int i = 0; i < N1; i++)
                {
                    InLayer.Neurons[i].Out = TestingSet[index, i];
                }

                for (int o = 0; o < N3; o++)
                {
                    Expected[o] = TestingSet[index, N1 + o];
                }
            } 
        }
        private double F(double x)
        {
            if (IsSigmoidFunction)
            {
                return (1.0 / (1 + Math.Exp(-x)));
            }
            else
            {
                return ((Math.Exp(x) - Math.Exp(-x)) / (Math.Exp(x) + Math.Exp(-x)));
            }
        }
        private double FDerived(double x)
        {
            if (IsSigmoidFunction)
            {
                return x * (1 - x);
            }
            else
            {
                double result = F(x);
                return (1 - Math.Pow(result, 2));
            }
        }
        private void Forward()
        {
            double sum;

            //propagate forward signal from input to hidden
            for (int h = 0; h < N2; h++)
            {
                sum = 0;
                for (int i = 0; i < N1; i++)
                {
                    sum += HidLayer.Neurons[h].W[i] * InLayer.Neurons[i].Out;
                }                         
                sum += HidLayer.Neurons[h].Bias;
                HidLayer.Neurons[h].Out = F(sum);
            }


            sum = 0;
            //propagate forward signal from hidden to output
            for (int o = 0; o < N3; o++)
            {
                sum = 0;
                for (int h = 0; h < N2; h++)
                {
                    sum += OutLayer.Neurons[o].W[h] * HidLayer.Neurons[h].Out;
                }
                sum += OutLayer.Neurons[o].Bias;
                OutLayer.Neurons[o].Out = F(sum);
            }
        }
        private double CheckError()
        {
            double cost = 0;
            for (int o = 0; o < N3; o++)
            {
                cost += Math.Pow(OutLayer.Neurons[o].Out - Expected[o], 2);
            }
            return cost / 2.0;
        }
        private void Backward()
        {
            double[] oError = new double[N3];
            double[] hError = new double[N2];
            double sum;
             
            for (int o = 0; o < N3; o++)
            {
                oError[o] = FDerived(OutLayer.Neurons[o].Out) * (OutLayer.Neurons[o].Out - Expected[o]);
            }

            for (int h = 0; h < N2; h++)
            {
                sum = 0;
                for (int o = 0; o < N3; o++)
                {
                    sum += oError[o] * OutLayer.Neurons[o].W[h];
                }
                hError[h] = FDerived(HidLayer.Neurons[h].Out) * sum;
            }

            for (int o = 0; o < N3; o++)
            {
                for (int h = 0; h < N2; h++)
                {
                    OutLayer.Neurons[o].W[h] = OutLayer.Neurons[o].W[h] - LearningRate * oError[o] * HidLayer.Neurons[h].Out;
                }
                OutLayer.Neurons[o].Bias = OutLayer.Neurons[o].Bias - LearningRate * oError[o];
            }
            for (int h = 0; h < N2; h++)
            {
                for (int i = 0; i < N1; i++)
                {
                    HidLayer.Neurons[h].W[i] = HidLayer.Neurons[h].W[i] - LearningRate * hError[h] * InLayer.Neurons[i].Out;
                }
                HidLayer.Neurons[h].Bias = HidLayer.Neurons[h].Bias - LearningRate * hError[h];
            }
        }
        public double Backpropagation()
        {
            double TrainingError = 0;
            for (int trainingData = 0; trainingData < TrainingSet.GetLength(0); trainingData++) // all training examples
            {
                SetData(trainingData, true);
                Forward();
                TrainingError += CheckError();
                Backward();
            }

            return TrainingError;
        }
        public double[] Predict(double[] inParams)
        {
            double[] outParams = new double[N3];
            int minVal;

            if (IsSigmoidFunction)
                minVal = 0;
            else
                minVal = -1;

            for (int i = 0; i < N1; i++)
            {
                inParams[i] = MathUtils.NormalizeData(inParams[i], Constants.cMinDataValue, Constants.cMaxDataValue, minVal, 1);
                InLayer.Neurons[i].Out = inParams[i];
            }
            Forward();
            for (int o = 0; o < N3; o++)
            {
                outParams[o] = OutLayer.Neurons[o].Out;
                outParams[o] = MathUtils.NormalizeData(outParams[o], minVal, 1, Constants.cMinDataValue, Constants.cMaxDataValue);
            }
            return outParams;
        }
        public double TestNetwork(string path)
        {
            var newFile = new FileInfo(path);
            ExcelWorksheet ws;
            double testingError = 0;
            string sheetName = "PredictionResults_N1+" + Convert.ToString(N2 - N1) + "_" + Convert.ToString(IsType1);
            using (ExcelPackage package = new ExcelPackage(newFile))
            {
                ws = package.Workbook.Worksheets.Add(sheetName);
                ws.Cells[1, 7].Value = "Learning Rate";
                ws.Cells[1, 8].Value = "Training error";
                ws.Cells[1, 9].Value = "Testing error";

                ws.Cells[1, 1].Value = "Time[h]";
                ws.Cells[1, 2].Value = "Predicted Alcohol Concentration";
                ws.Cells[1, 3].Value = "Expected Alcohol Concentration";
                ws.Cells[1, 4].Value = "Predicted Substrate";
                ws.Cells[1, 5].Value = "Expected Substrate";

                int minVal;

                if (IsSigmoidFunction)
                {
                    minVal = 0;
                }
                else
                {
                    minVal = -1;
                }
                for (int testData = 0; testData < TestingSet.GetLength(0); testData++)
                {
                    SetData(testData, false);
                    Forward();
                    testingError += CheckError();
                    ws.Cells[testData + 2, 1].Value = Math.Round(MathUtils.NormalizeData(InLayer.Neurons[1].Out, minVal, 1, Constants.cMinDataValue, Constants.cMaxDataValue));
                    ws.Cells[testData + 2, 2].Value = MathUtils.NormalizeData(OutLayer.Neurons[0].Out, minVal, 1, Constants.cMinDataValue, Constants.cMaxDataValue);
                    ws.Cells[testData + 2, 3].Value = MathUtils.NormalizeData(Expected[0], minVal, 1, Constants.cMinDataValue, Constants.cMaxDataValue);
                    ws.Cells[testData + 2, 4].Value = MathUtils.NormalizeData(OutLayer.Neurons[1].Out, minVal, 1, Constants.cMinDataValue, Constants.cMaxDataValue);
                    ws.Cells[testData + 2, 5].Value = MathUtils.NormalizeData(Expected[1], minVal, 1, Constants.cMinDataValue, Constants.cMaxDataValue);
                
                }
                ws.Cells[2, 7].Value = LearningRate;
                ws.Cells[2, 8].Value = TrainingError;
                ws.Cells[2, 9].Value = testingError;
                package.Save();
            }
            return testingError;
        }
    }
}
