﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using System.Threading;
using WineFermentationAITool.LibFiles;
using System.IO;
using Newtonsoft.Json;

namespace WineFermentationAITool.LibFiles
{
    public class PSO
    {
        List<Particle> Particles;
        int NumOfParticles;
        double c1; /* cognitive coefficient */
        double c2; /* social coefficient */
        double w;  /* inertia weight */
        double R1;
        double R2;
        public int N2;
        int N1, N3;
        int NumOfIterations;
        int MaxEpochsPSO;
        Chromosome GlobalBest;
        NeuralNetwork PSONeuralNetwork;
        List<Particle> ParticlesFitness;
        public bool IsType1;
        public bool IsSigmoid;

        void SetWeights(Particle Particle)
        {
            double[] W;

            for (int h = 0; h < N2; h++)
            {
                W = Particle.Position.W12.Cast<double>().ToArray();
                Array.Copy(W, h * N1, PSONeuralNetwork.HidLayer.Neurons[h].W, 0, N1);
                PSONeuralNetwork.HidLayer.Neurons[h].Bias = Particle.Position.HBias[h];
            }

            for (int o = 0; o < N3; o++)
            {
                W = Particle.Position.W23.Cast<double>().ToArray();
                Array.Copy(W, o * N2, PSONeuralNetwork.OutLayer.Neurons[o].W, 0, N2);
                PSONeuralNetwork.OutLayer.Neurons[o].Bias = Particle.Position.HBias[o];
            }
        }
        public void GetParticleFitness(Particle particle)
        {
            SetWeights(particle);
            int CurrentIteration = 0;

            while (CurrentIteration++ < MaxEpochsPSO)
            {
                particle.Fitness = PSONeuralNetwork.Backpropagation();
            }

            if (particle.Fitness < particle.PersonalBest.Fitness)
            {
                particle.PersonalBest = particle.Position;
            }
            if (particle.Fitness < GlobalBest.Fitness)
            {
                GlobalBest = particle.Position;
            }
            ParticlesFitness.Add(particle);
        }

        public PSO(int numOfParticles, double c1, double c2, double w, int numOfIterations, int maxEpochsPSO, int N1, int N2, int N3, double learningRate, bool isSigmoidFunction, bool isType1, bool[] isTrainingSet, List<double[,]> dataSets, string weightsAction, string wFileName)
        {
            Random rand = new Random();
            this.NumOfParticles = numOfParticles;
            Particles = new List<Particle>();
            this.c1 = c1;
            this.c2 = c2;
            this.w = w;
            this.R1 = rand.NextDouble();
            this.R2 = rand.NextDouble();
            this.NumOfIterations = numOfIterations;
            this.MaxEpochsPSO = maxEpochsPSO;
            this.N1 = N1;
            this.N2 = N2;
            this.N3 = N3; 
            this.ParticlesFitness = new List<Particle>();
            this.IsType1 = isType1;
            this.IsSigmoid = isSigmoidFunction;
            this.GlobalBest = new();
            GlobalBest.Fitness = Double.MaxValue;
            /* evaluate all particles and select the global best */
            //List<Task> tasks = new();

            if (weightsAction == Constants.LOAD)
            {
                using (StreamReader sr = new StreamReader(wFileName))
                {
                    this.IsType1 = Convert.ToBoolean(sr.ReadLine().Split(':')[1]);
                    this.N2 = Convert.ToInt32(sr.ReadLine().Split(':')[1]);
                    this.IsSigmoid = Convert.ToBoolean(sr.ReadLine().Split(':')[1]);
                    NumOfParticles = Convert.ToInt32(sr.ReadLine().Split(':')[1]);

                    for (int particleNo = 0; particleNo < NumOfParticles; particleNo++)
                    {
                        var particleString = sr.ReadLine();
                        Chromosome position = particleString == null ? new() : JsonConvert.DeserializeObject<Chromosome>(particleString);
                        Particle particle = new Particle(N1, N2, N3); 
                        particle.Position = new Chromosome(position);
                        Particles.Add(particle);
                    }
                }
            }
            else
            {
                for (int particleNo = 0; particleNo < NumOfParticles; particleNo++)
                {
                    Particle particle = new Particle(N1, N2, N3);
                    Particles.Add(particle);
                }
            }
            
            if (weightsAction == Constants.SAVE)
            {
                using (StreamWriter sw = new StreamWriter(wFileName))
                {
                    sw.WriteLine("IsType1:" + Convert.ToString(IsType1));
                    sw.WriteLine("N2:" + Convert.ToString(N2));
                    sw.WriteLine("IsSigmoid:" + Convert.ToString(IsSigmoid));
                    sw.WriteLine("PopulationSize:" + Convert.ToString(NumOfParticles));

                    foreach (var particle in Particles)
                    {
                        var particleString = JsonConvert.SerializeObject(particle.Position);
                        sw.WriteLine(particleString);
                    }
                }
            }
            
            
            this.PSONeuralNetwork = new NeuralNetwork(isType1, N2 - N1, learningRate * 10, isSigmoidFunction, "xavier", maxEpochsPSO, isTrainingSet, dataSets);

            var threads_counter = 0;
            List<Task> tasks = new();
            foreach (var particle in Particles)
            {
                if (threads_counter == Constants.MAX_NUM_OF_THREADS)
                {
                    /* Wait for all the 8 tasks to finish and clear them from the list */
                    Task.WaitAll(tasks.ToArray());
                    tasks.Clear();
                    threads_counter = 0;
                }
                tasks.Add(Task.Factory.StartNew(() => GetParticleFitness(particle)));
            }
            Task.WaitAll(tasks.ToArray());
            tasks.Clear();

            Particles = new(ParticlesFitness);
            ParticlesFitness.Clear();
        }

        public void getIndividualsFitness(int particleNo)
        {
            for (int i = 0; i < N2; i++)
            {
                for (int j = 0; j < N1; j++)
                {
                    Particles[particleNo].Velocity.W12[i, j] *= w;
                    Particles[particleNo].Velocity.W12[i, j] += c1 * R1 * (Particles[particleNo].PersonalBest.W12[i, j] - Particles[particleNo].Position.W12[i, j]);
                    Particles[particleNo].Velocity.W12[i, j] += c2 * R2 * (GlobalBest.W12[i, j] - Particles[particleNo].Position.W12[i, j]);
                    Particles[particleNo].Position.W12[i, j] *= Particles[particleNo].Velocity.W12[i, j];
                }
                Particles[particleNo].Velocity.HBias[i] *= w;
                Particles[particleNo].Velocity.HBias[i] += c1 * R1 * (Particles[particleNo].PersonalBest.HBias[i] - Particles[particleNo].Position.HBias[i]);
                Particles[particleNo].Velocity.HBias[i] += c2 * R2 * (GlobalBest.HBias[i] - Particles[particleNo].Position.HBias[i]);
                Particles[particleNo].Position.HBias[i] *= Particles[particleNo].Velocity.HBias[i];
            }

            for (int i = 0; i < N3; i++)
            {
                for (int j = 0; j < N2; j++)
                {
                    Particles[particleNo].Velocity.W23[i, j] *= w;
                    Particles[particleNo].Velocity.W23[i, j] += c1 * R1 * (Particles[particleNo].PersonalBest.W23[i, j] - Particles[particleNo].Position.W23[i, j]);
                    Particles[particleNo].Velocity.W23[i, j] += c2 * R2 * (GlobalBest.W23[i, j] - Particles[particleNo].Position.W23[i, j]);
                    Particles[particleNo].Position.W23[i, j] *= Particles[particleNo].Velocity.W23[i, j];
                }
                Particles[particleNo].Velocity.OBias[i] *= w;
                Particles[particleNo].Velocity.OBias[i] += c1 * R1 * (Particles[particleNo].PersonalBest.OBias[i] - Particles[particleNo].Position.OBias[i]);
                Particles[particleNo].Velocity.OBias[i] += c2 * R2 * (GlobalBest.OBias[i] - Particles[particleNo].Position.OBias[i]);
                Particles[particleNo].Position.OBias[i] *= Particles[particleNo].Velocity.OBias[i];
            }
            GetParticleFitness(Particles[particleNo]);
        }
        public Chromosome PSOOptimization()
        {
            int iterationNo = 0;

            while (iterationNo++ < NumOfIterations)
            {
                for (int particleNo = 0; particleNo < NumOfParticles; particleNo++)
                {
                    getIndividualsFitness(particleNo);
                }
                Particles = new(ParticlesFitness);
                ParticlesFitness.Clear();
            }

            return GlobalBest;
        }
    }
}
