using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Mvc.RazorPages;
using WineFermentationAITool.Models;
using WineFermentationAITool.LibFiles;
using System.IO;
using Newtonsoft.Json;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Http;

namespace WineFermentationAITool.Pages.Forms
{
    public class PredictParamsFormModel : PageModel
    {
        [BindProperty]
        public PredictionModel Pred_Model { get; set; }

        [BindProperty]
        public string FileName { get; set; }

        [BindProperty]
        public bool IsDefaultConfiguration { get; set; }
        [BindProperty]
        public IFormFile UploadFiles { get; set; }
        [BindProperty]
        public bool IsType1 { get; set; }

        //  private readonly IWebHostEnvironment _he;
        public PredictParamsFormModel(IWebHostEnvironment he)
        {
          //  _he = he;
        }

        public void OnGet()
        {
            var strIsDefaultConfiguration = HttpContext.Session.GetString("IsDefaultConfiguration");
            IsDefaultConfiguration = strIsDefaultConfiguration == null ? true : JsonConvert.DeserializeObject<bool>(strIsDefaultConfiguration.ToLower());
            var strPredictionModel = HttpContext.Session.GetString("Pred_Model");
            Pred_Model = strPredictionModel == null ? default(PredictionModel) : JsonConvert.DeserializeObject<PredictionModel>(strPredictionModel);
            FileName = HttpContext.Session.GetString("FileName");
            if (Pred_Model != null)
            {
                Pred_Model.alcoholConcentration = Math.Round(Pred_Model.alcoholConcentration, 4);
                Pred_Model.substrateConcentration = Math.Round(Pred_Model.substrateConcentration, 4);
            }
        }
        public async Task<IActionResult> OnPost()
        { 
            double[,] W12;
            double[,] W23;
            double[] HBias;
            double[,] OBias;
            bool IsType1, IsSigmoid;
            int N2, PopulationSize;
            Chromosome configuration;
            double[] results;
            string FileNameWithPath="";

            if (IsDefaultConfiguration && FileName == null)
            {
                FileName = "Default_NN_Configuration.txt";
               // FileNameWithPath = Path.Combine(_he.ContentRootPath, "Files", FileName);
                FileNameWithPath = Path.Combine(@"D:\Work\MAPIoT\WineFermentationSite\","Files", FileName);
            }


            if (!IsDefaultConfiguration)
            {
                FileName = UploadFiles.FileName; 
               // FileNameWithPath = Path.Combine(_he.ContentRootPath, "Files", FileName);
                FileNameWithPath = Path.Combine(@"D:\Work\MAPIoT\WineFermentationSite\", "Files", FileName);
                using (var fs = new FileStream(FileNameWithPath, FileMode.Open))
                {
                    await UploadFiles.CopyToAsync(fs);
                    ViewData["Message"] = "The Selected File " + UploadFiles.FileName + " Is Uploaded Successfully";
                }
            }
          

            //readFile
            using (StreamReader sr = new StreamReader(FileNameWithPath))
            {
                IsType1 = Convert.ToBoolean(sr.ReadLine().Split(':')[1]);
                N2 = Convert.ToInt32(sr.ReadLine().Split(':')[1]);
                IsSigmoid = Convert.ToBoolean(sr.ReadLine().Split(':')[1]);
                PopulationSize = Convert.ToInt32(sr.ReadLine().Split(':')[1]);
                var configurationString = sr.ReadLine();
                configuration = configurationString == null ? new() : JsonConvert.DeserializeObject<Chromosome>(configurationString);
            }

            NeuralNetwork MyNeuralNetwork = new NeuralNetwork(IsType1, N2, IsSigmoid, configuration.W12, configuration.W23, configuration.HBias, configuration.OBias);
            results = MyNeuralNetwork.Predict(new double[]{ Pred_Model.biomassConcentration, Pred_Model.time, Pred_Model.temperature, Pred_Model.initialSubstrate });

            Pred_Model.alcoholConcentration = results[0];
            Pred_Model.substrateConcentration = results[1];
            HttpContext.Session.SetString("Pred_Model", JsonConvert.SerializeObject(Pred_Model));
            HttpContext.Session.SetString("IsDefaultConfiguration", Convert.ToString(IsDefaultConfiguration));
            HttpContext.Session.SetString("FileName", Convert.ToString(FileName));
            return RedirectToPage("/Forms/PredictParamsForm");
        }
    }
}
