using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Mvc.RazorPages;
using Microsoft.AspNetCore.Http;
using WineFermentationAITool.Models;
using OfficeOpenXml;
using Microsoft.AspNetCore.Hosting;
using System.IO;
using Newtonsoft.Json;
using WineFermentationAITool.LibFiles;

namespace WineFermentationAITool.Pages.Forms
{
    public class ResultsFormModel : PageModel
    {
        [BindProperty]
        public double trainingError { get; set; }
        [BindProperty]
        public double testingError { get; set; }
        double[,] testingSet { get; set; }
        bool isType1 { get; set; }
        bool isSigmoid { get; set; }
        public string GoogleDataTableJson { get; private set; }

        private List<ResultsChart> predictedAlcoholList;
        private List<ResultsChart> expectedAlcoholList;
        private List<ResultsChart> predictedSubstrateList;
        private List<ResultsChart> expectedSubstrateList;
        List<double> biomass;
        List<double> alcohol;
        List<double> substrate;

      //  private readonly IWebHostEnvironment _he;
        public ResultsFormModel(IWebHostEnvironment he)
        {
         //   _he = he;
            biomass = new();
            alcohol = new();
            substrate = new();
        }

        //public FileResult Download()
        public IActionResult OnGet()
        {

            var FileResults = HttpContext.Session.GetString("FileResults");
            //string path = Path.Combine(_he.ContentRootPath, "Files", "Results.xlsx");
            //string path = Path.Combine(@"D:\Work\MAPIoT\WineFermentationSite\", "Files", "Results.xlsx");
            byte[] bytes = System.IO.File.ReadAllBytes(FileResults);
            if (Request.Headers["Referer"].ToString() == "")
            {
                return RedirectToPage("/Index");
            }
            var NN_String = HttpContext.Session.GetString("NN_Model");
            NeuralNetworkModel NN_Model = NN_String == null ? new() : JsonConvert.DeserializeObject<NeuralNetworkModel>(NN_String);
            trainingError = Math.Round(NN_Model.TrainingError, 4);
            testingError = Math.Round(NN_Model.TestingError,4);
            return Page();
            //return File(bytes, "application/octet-stream", "Results.xlsx");
        }

        public void ReadChartData()
        {

            var FilePath = HttpContext.Session.GetString("FileResults");
            ExcelPackage.LicenseContext = LicenseContext.NonCommercial; 
            //string FilePath = Path.Combine(_he.ContentRootPath, "Files", "Results.xlsx");
            //string FilePath = Path.Combine(@"D:\Work\MAPIoT\WineFermentationSite\", "Files", "Results.xlsx");
            ExcelCellAddress start,
                             end;
            using var fs = new FileStream(FilePath, FileMode.Open);
            using var package = new ExcelPackage(fs);
            predictedAlcoholList = new List<ResultsChart>();
            expectedAlcoholList = new List<ResultsChart>();
            predictedSubstrateList = new List<ResultsChart>();
            expectedSubstrateList = new List<ResultsChart>();

            foreach (var Worksheet in package.Workbook.Worksheets)
            {
                start = Worksheet.Dimension.Start;
                end = Worksheet.Dimension.End;

                for (int row = start.Row + 1; row <= end.Row; row++)
                {
                    predictedAlcoholList.Add(new ResultsChart
                    { 
                        time = Convert.ToInt32(Worksheet.Cells[row, 1].Value),
                        value = Convert.ToDouble(Worksheet.Cells[row, 2].Value)
                    });
                    expectedAlcoholList.Add(new ResultsChart
                    {
                        time = Convert.ToInt32(Worksheet.Cells[row, 1].Value),
                        value = Convert.ToDouble(Worksheet.Cells[row, 3].Value)
                    });
                    predictedSubstrateList.Add(new ResultsChart
                    {
                        time = Convert.ToInt32(Worksheet.Cells[row, 1].Value),
                        value = Convert.ToDouble(Worksheet.Cells[row, 4].Value)
                    });
                    expectedSubstrateList.Add(new ResultsChart
                    {
                        time = Convert.ToInt32(Worksheet.Cells[row, 1].Value),
                        value = Convert.ToDouble(Worksheet.Cells[row, 5].Value)
                    });
                    alcohol.Add(Convert.ToDouble(Worksheet.Cells[row, 2].Value));
                    substrate.Add(Convert.ToDouble(Worksheet.Cells[row, 4].Value));
                }
            }

            ReadInputData();
            ApplyRules();
        }

        public void ReadInputData()
        {
            var NN_String = HttpContext.Session.GetString("NN_Model");
            NeuralNetworkModel NN_Model = NN_String == null ? new() : JsonConvert.DeserializeObject<NeuralNetworkModel>(NN_String);
            isType1 = NN_Model.IsType1;
            isSigmoid = NN_Model.IsSigmoid;
            testingSet = NN_Model.TestingSet;

            int minVal = -1;
            if (isSigmoid)
            {
                minVal = 0;
            }

            if (isType1)
            {
                for (var row = 0; row < testingSet.GetLength(0); row++)
                {
                    var value = MathUtils.NormalizeData(testingSet[row, 0], minVal, 1, Constants.cMinDataValue, Constants.cMaxDataValue);
                    biomass.Add(value);
                }
            }
            else
            {

            }
        }

        bool isIncreasing(double currVal, double prevVal, double treshold)
        {
            if (currVal - prevVal > treshold)
                return true;
            return false;
        }
        bool isConstant(double currVal, double prevVal, double treshold)
        {
            if (prevVal - currVal < treshold)
                return true;
            return false;
        }
        public void ApplyRules()
        {
            string lastPhase = "latent";
            double minAlcohol = alcohol.Min();
            for (int i = 0; i < biomass.Count; i++)
            {
                if (biomass[i] < 0.2 && alcohol[i] < (minAlcohol + 1) && ((i > 0) && (isConstant(substrate[i], substrate[i - 1], 0.5)) || substrate[i] > 205) && lastPhase == "latent")
                {
                    predictedAlcoholList[i].tooltip = "Latent phase: \n" + Convert.ToString(Math.Round(predictedAlcoholList[i].value, 4));
                    predictedSubstrateList[i].tooltip = "Latent phase: \n" + Convert.ToString(Math.Round(predictedSubstrateList[i].value, 4));
                    lastPhase = "latent";
                }
                else if (biomass[i] > 0.2 && isIncreasing(alcohol[i], alcohol[i - 1], 0.5) && substrate[i] > 20 && (lastPhase == "latent" || lastPhase == "exponential"))
                {
                    predictedAlcoholList[i].tooltip = "Exponential growth phase: \n" + Convert.ToString(Math.Round(predictedAlcoholList[i].value, 4));
                    predictedSubstrateList[i].tooltip = "Exponential growth phase: \n" + Convert.ToString(Math.Round(predictedSubstrateList[i].value, 4));
                    lastPhase = "exponential";
                }
                else if (biomass[i] < 2.2 && isConstant(alcohol[i], alcohol[i - 1], 2) && substrate[i] < 40 && (lastPhase == "decay" || lastPhase == "exponential"))
                {
                    predictedAlcoholList[i].tooltip = "Decay phase: \n" + Convert.ToString(Math.Round(predictedAlcoholList[i].value, 4));
                    predictedSubstrateList[i].tooltip = "Decay phase: \n" + Convert.ToString(Math.Round(predictedSubstrateList[i].value, 4));
                    lastPhase = "decay";
                }
                else
                {
                    predictedAlcoholList[i].tooltip = "Undefined: \n" + Convert.ToString(Math.Round(predictedAlcoholList[i].value, 4));
                    predictedSubstrateList[i].tooltip = "Undefined: \n" + Convert.ToString(Math.Round(predictedSubstrateList[i].value, 4));
                }
            }
        }

        public JsonResult OnGetChartData()
        {
            ReadChartData();
            List<List<ResultsChart>> chartData = new()
            {
                predictedAlcoholList,
                expectedAlcoholList,
                predictedSubstrateList,
                expectedSubstrateList
            };
            return new JsonResult(chartData);
        }
    }
}
