using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Mvc.RazorPages;
using System.IO;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Http;
using Newtonsoft.Json;
using OfficeOpenXml;
using WineFermentationAITool.Models;
using WineFermentationAITool.LibFiles;

namespace WineFermentationAITool.Pages.Forms
{
    public class TrainTestFormModel : PageModel
    {
        [BindProperty]
        public NeuralNetworkModel NN_Model { get; set; }
        public GeneticAlgorithmModel GA_Model { get; set; }
        public ParticleSwarmOptimizationModel PSO_Model { get; set; }
        [BindProperty]
        public IFormFile UploadFiles { get; set; }
        [BindProperty]
        public bool SaveNNConfig { get; set; }
        [BindProperty]
        public string ConfigFileName { get; set; }
        public double Percentage { get; private set; }

       // private readonly IWebHostEnvironment _he;

        public TrainTestFormModel(IWebHostEnvironment he)
        {
          //  _he = he;
        }

        public IActionResult OnGet()
        {
            if (Request.Headers["Referer"].ToString() == "")
            {
                return RedirectToPage("/Index");
            }
            var NN_String = HttpContext.Session.GetString("NN_Model");
            NN_Model = NN_String == null ? new() : JsonConvert.DeserializeObject<NeuralNetworkModel>(NN_String);

            if (NN_Model.IsTrainingSet == null)
            {
                NN_Model.IsTrainingSet = new bool[4];
            }
            if (NN_Model.DataSets == null)
            {
                NN_Model.DataSets = new List<double[,]>();
            }
            return Page();
        }

        public void ReadDataFile(string filePath)
        {
            ExcelPackage.LicenseContext = LicenseContext.NonCommercial;
            ExcelCellAddress start,
                             end;
            double[,] DataSet;
            
            int lowerBound, 
                upperBound;

            if (NN_Model.IsSigmoid)
            {
                lowerBound = 0;
                upperBound = 1;
            } else
            {
                lowerBound = -1;
                upperBound = 1;
            }

            using (var package = new ExcelPackage(filePath))
            {
                foreach (var Worksheet in package.Workbook.Worksheets)
                {
                    start = Worksheet.Dimension.Start;
                    end = Worksheet.Dimension.End;
                    DataSet = new double[end.Row - 1, end.Column];

                    for (int row = start.Row + 1; row <= end.Row; row++)
                    {
                        for (int col = start.Column; col <= end.Column; col++)
                        {
                            DataSet[row - 2, col - 1] = Convert.ToDouble(Worksheet.Cells[row, col].Text);
                            DataSet[row - 2, col - 1] = MathUtils.NormalizeData(DataSet[row - 2, col - 1], Constants.cMinDataValue, Constants.cMaxDataValue, lowerBound, upperBound);
                        }
                    }
                    NN_Model.DataSets.Add(DataSet);
                }
            }
        }
        public async Task<IActionResult> OnPostTrainTestAsync()
        {

            var NN_string = HttpContext.Session.GetString("NN_Model");
            var GA_string = HttpContext.Session.GetString("GA_Model");
            var PSO_string = HttpContext.Session.GetString("PSO_Model");

            var NN_Temp_Model = NN_string == null ? default(NeuralNetworkModel) : JsonConvert.DeserializeObject<NeuralNetworkModel>(NN_string);
            GA_Model = GA_string == null ? default(GeneticAlgorithmModel) : JsonConvert.DeserializeObject<GeneticAlgorithmModel>(GA_string);
            PSO_Model = PSO_string == null ? default(ParticleSwarmOptimizationModel) : JsonConvert.DeserializeObject<ParticleSwarmOptimizationModel>(PSO_string);
            NN_Model.DataSets = new();

            //var FileUpload = Path.Combine(_he.ContentRootPath, "Files", UploadFiles.FileName);
            var FileUpload = Path.Combine(@"D:\Work\MAPIoT\WineFermentationSite\", "Files", UploadFiles.FileName);  
            /*string msg = "<script language=\"javascript\">";
            msg += "alert('" + FileUpload.ToString() + "');";
            msg += "</script>";
            Response.WriteAsync(msg);
            */
            using (var fs = new FileStream(FileUpload, FileMode.Create))
            {
                await UploadFiles.CopyToAsync(fs);
                ViewData["Message"] = "The Selected File " + UploadFiles.FileName + " Is Uploaded Successfully";
            }
            
            NN_Model.IsSigmoid = NN_Temp_Model.IsSigmoid;
            ReadDataFile(FileUpload);
            NN_Temp_Model.IsTrainingSet = NN_Model.IsTrainingSet;
            NN_Temp_Model.DataSets = NN_Model.DataSets;
            NN_Model = NN_Temp_Model;
                
            TrainAndTest();
            HttpContext.Session.SetString("NN_Model", JsonConvert.SerializeObject(NN_Model));
            HttpContext.Session.SetString("DataSetFileName", JsonConvert.SerializeObject(UploadFiles.FileName));
            return RedirectToPage("/Forms/ResultsForm");
        }
        
        public void TrainAndTest()
        {
            var fileCt = 0;
            string tempResults;
            var Results = "Results_" +
                          (NN_Model.IsType1 ? "Type1_" : "Type2_") +
                          Convert.ToString(NN_Model.HiddenLayer) + "_" +
                          (NN_Model.IsSigmoid ? "Sigmoid_" : "Tanh_") +
                          Convert.ToString(NN_Model.LearningRate) + "_" +
                          NN_Model.InitType + ".xlsx";
            //var FileResults = Path.Combine(_he.ContentRootPath, "Files", "Results.xlsx");
            var FileResults = Path.Combine(@"D:\Work\MAPIoT\WineFermentationSite\", "Files", Results);
         /*   string msg = "<script language=\"javascript\">";
            msg += "alert('" + @FileResults + "');";
            msg += "</script>";
            Response.WriteAsync(msg);
        */
            while (System.IO.File.Exists(FileResults))
            {
                fileCt++;
                string[] split = Results.Split('.');
                tempResults = string.Join(".", split.Take(split.Length - 1));
                tempResults += "(" + Convert.ToString(fileCt) + ").xlsx";
                FileResults = Path.Combine(@"D:\Work\MAPIoT\WineFermentationSite\", "Files", tempResults);
            }

            HttpContext.Session.SetString("FileResults", FileResults);

            NeuralNetwork MyNeuralNetwork = new NeuralNetwork(NN_Model.IsType1, NN_Model.HiddenLayer, NN_Model.LearningRate, NN_Model.IsSigmoid, NN_Model.InitType, NN_Model.MaxEpochs, NN_Model.IsTrainingSet, NN_Model.DataSets);
            if (NN_Model.InitType == Constants.GA)
            {
                string weightsFileNameWithPath = "";
                if (GA_Model.weightsAction is Constants.SAVE or Constants.LOAD)
                {
                    //weightsFileNameWithPath = Path.Combine(_he.ContentRootPath, "Files", GA_Model.weightsFileName);
                    weightsFileNameWithPath = Path.Combine(@"D:\Work\MAPIoT\WineFermentationSite\", "Files", GA_Model.weightsFileName);
                }
                MyNeuralNetwork.SetGAParams(GA_Model.populationSize, GA_Model.maxEpochs, GA_Model.maxGenerations, GA_Model.pCrossover, GA_Model.pMutation, GA_Model.Beta, GA_Model.isElitist, NN_Model.IsTrainingSet, NN_Model.DataSets, GA_Model.weightsAction, weightsFileNameWithPath);
            }
            else if (NN_Model.InitType == Constants.PSO)
            {
                string weightsFileNameWithPath = "";
                if (PSO_Model.weightsAction is Constants.SAVE or Constants.LOAD)
                {
                   // weightsFileNameWithPath = Path.Combine(_he.ContentRootPath, "Files", PSO_Model.weightsFileName);
                    weightsFileNameWithPath = Path.Combine(@"D:\Work\MAPIoT\WineFermentationSite\", "Files", PSO_Model.weightsFileName);
                }

                MyNeuralNetwork.SetPSOParams(PSO_Model.numOfParticles, PSO_Model.c1, PSO_Model.c2, PSO_Model.w, PSO_Model.numOfIterations, PSO_Model.maxEpochsPSO, NN_Model.IsTrainingSet, NN_Model.DataSets, PSO_Model.weightsAction, weightsFileNameWithPath);
            }

            int CurrentIteration = 0;
            NN_Model.TrainingError = double.MaxValue;
            while (NN_Model.TrainingError > Math.Pow(10, -5) && CurrentIteration < NN_Model.MaxEpochs)
            {
                NN_Model.TrainingError = MyNeuralNetwork.Backpropagation();
                Percentage = (double)CurrentIteration / NN_Model.MaxEpochs * 100;
                HttpContext.Session.SetString("Percentage", Convert.ToString(Percentage));
                CurrentIteration++;
            }
            if (SaveNNConfig)
            {
                //var FileNameWithPath = Path.Combine(_he.ContentRootPath, "Files", ConfigFileName);
                var FileNameWithPath = Path.Combine(@"D:\Work\MAPIoT\WineFermentationSite\", "Files", ConfigFileName);
                using (StreamWriter sw = new StreamWriter(FileNameWithPath))
                {
                    sw.WriteLine("IsType1:" + Convert.ToString(NN_Model.IsType1));
                    sw.WriteLine("N2:" + Convert.ToString(NN_Model.HiddenLayer));
                    sw.WriteLine("IsSigmoid:" + Convert.ToString(NN_Model.IsSigmoid));
                    sw.WriteLine("PopulationSize:" + Convert.ToString(0));

                    Chromosome configuration = MyNeuralNetwork.GetNeuralNetworkConfiguration();
                    var configurationString = JsonConvert.SerializeObject(configuration);
                    sw.WriteLine(configurationString);
                }
            }

            NN_Model.TestingError = MyNeuralNetwork.TestNetwork(FileResults);
            NN_Model.TestingSet = MyNeuralNetwork.GetTestingSet();
        }

        public IActionResult OnPostPrevious()
        {
            // restore values received from neuralNetworkForm
            var NN_String = HttpContext.Session.GetString("NN_Model");
            var NN_Temp_Model = NN_String
                == null ? new() : JsonConvert.DeserializeObject<NeuralNetworkModel>(NN_String);

            NN_Temp_Model.IsTrainingSet = NN_Model.IsTrainingSet;
            NN_Temp_Model.DataSets = NN_Model.DataSets;
            HttpContext.Session.SetString("NN_Model", JsonConvert.SerializeObject(NN_Temp_Model));

            if (NN_Temp_Model.InitType == Constants.XAVIER)
            {
                return RedirectToPage("/Forms/NeuralNetworkForm");
            }
            else if (NN_Temp_Model.InitType == Constants.GA)
            {
                return RedirectToPage("/Forms/GeneticAlgorithmForm");
            }
            else if (NN_Temp_Model.InitType == Constants.PSO)
            {
                return RedirectToPage("/Forms/ParticleSwarmOptimizationForm");
            }
            else
            {
                //TODO: add error page
                return RedirectToPage("");
            }
        }
    }
}
